/*****************************************************************************
 * libsangoma.c	AFT T1/E1: HDLC API Code Library
 *
 * Author(s):	Anthony Minessale II <anthmct@yahoo.com>
 *              Nenad Corbic <ncorbic@sangoma.com>
 *
 * Copyright:	(c) 2005 Anthony Minessale II
 *
 *		This program is free software; you can redistribute it and/or
 *		modify it under the terms of the GNU General Public License
 *		as published by the Free Software Foundation; either version
 *		2 of the License, or (at your option) any later version.
 * ============================================================================
 */

#include "libsangoma.h"

#define DFT_CARD "wanpipe1"


int sangoma_interface_toi(char *interface, int *span, int *chan)
{
	char *data, *p, *sp = NULL, *ch = NULL;
	int ret = 0;
	
	if ((data = strdupa(interface))) {
		for (p = data; *p; p++) {
			if (sp && *p == 'g') {
				*p = '\0';
				ch = (p + 1);
				break;
			} else if (*p == 'w') {
				sp = (p + 1);
			}
		}

		if(ch && sp) {
			*span = atoi(sp);
			*chan = atoi(ch);
			ret = 1;
		} else {
			*span = -1;
			*chan = -1;
		}
	}

	return ret;
}

int sangoma_create_socket_by_name(char *device, char *card) 
{
	struct wan_sockaddr_ll  sa;
	int sock=-1;

	if (!card) {
		card = DFT_CARD;
	}
	memset(&sa, 0, sizeof(struct wan_sockaddr_ll));
	sock = socket(AF_WANPIPE, SOCK_RAW, 0);
	if ( sock >= 0 ) {
		strncpy(sa.sll_device, device, sizeof(sa.sll_device));
		strncpy(sa.sll_card, card ? card : DFT_CARD, sizeof(sa.sll_card));
        sa.sll_protocol = htons(PVC_PROT);
        sa.sll_family = AF_WANPIPE;

        if (bind(sock, (struct sockaddr *)&sa, sizeof(struct wan_sockaddr_ll)) < 0) {
			close(sock);
			sock = -1;
        }

	} 


	return sock;
}

int sangoma_create_socket_intr(int span, int intr) 
{
	char device[25];
	sprintf(device, "w%dg%d", span, intr);
	return sangoma_create_socket_by_name(device, DFT_CARD);
}


int sangoma_create_socket(int span) 
{
	int sock=-1;
	int i;

	for (i=1; i<32; i++) {
		if (i == 16 || i == 24) {
			continue;
		}
		if ((sock=sangoma_create_socket_intr(span, i) >= 0)) {
			break;
		}
	}

	return sock;

}


int sangoma_write_socket(int sock, void *data, int len)
{

	return send(sock,data,len,0);
}


/*==============================================================
 * sangoma_sendmsg_socket
 * 
 * Transmits wanpipe header and data using two separate buffers
 *
 * INPUTS: 
 *   
 *  hdrbuf  = header buffer must be size of sizeof(api_rx_hdr_t) 16bytes
 *  hdrlen  = lenght of the header buffer sizeof(api_rx_hdr_t) 16bytes
 *   
 *  databuf = data buffer that will contain the actual payload
 *  datalen = length of the data buffer. Always set it to MAX value.
 *  
 * OUTPUT:
 * 
 *  hdrbuf  = will contain Sangoma 16 byte header
 *  databuf = will contain data payload
 *  
 *  Function returns length of received databuf WITHOUT the header.
 *  
 */


int sangoma_sendmsg_socket(int sock, void *hdrbuf, int hdrlen, void *databuf, int datalen, int flag)
{
	int bsent;
	struct msghdr msg;
	struct iovec iov[2];

	memset(&msg,0,sizeof(struct msghdr));

	iov[0].iov_len=hdrlen;
	iov[0].iov_base=hdrbuf;

	iov[1].iov_len=datalen;
	iov[1].iov_base=databuf;

	msg.msg_iovlen=2;
	msg.msg_iov=iov;

	bsent = sendmsg(sock,&msg,flag);
	if (bsent > 0){
		bsent-=sizeof(sangoma_api_hdr_t);
	}

	return bsent;
}



int sangoma_read_socket(int sock, void *data, int len)
{
	api_rx_element_t *api_rx_el;
	int rx_len = recv(sock,data,len,0);

	if (rx_len <= sizeof(api_rx_hdr_t)){
		return -EINVAL;
	}

	api_rx_el = (api_rx_element_t*)data;


	if (api_rx_el->api_rx_hdr.error_flag){
		return -EINVAL;
	}

	return rx_len;
}

/*==============================================================
 * sangoma_readmsg_socket
 *
 * Receives wanpipe header and data in two separate buffers
 *
 * INPUTS: 
 * 
 * hdrbuf  = header buffer must be size of sizeof(api_rx_hdr_t) 16bytes
 * hdrlen  = lenght of the header buffer sizeof(api_rx_hdr_t) 16bytes
 *
 * databuf = data buffer that will contain the actual payload
 * datalen = length of the data buffer. Always set it to MAX value.
 *
 * OUTPUT:
 *
 * hdrbuf  = will contain Sangoma 16 byte header
 * databuf = will contain data payload
 *
 * Function returns length of received databuf WITHOUT the header.
 *
 */

int sangoma_readmsg_socket(int sock, void *hdrbuf, int hdrlen, void *databuf, int datalen, int flag)
{
	int rx_len;
	struct msghdr msg;
	struct iovec iov[2];

	memset(&msg,0,sizeof(struct msghdr));

	iov[0].iov_len=hdrlen;
	iov[0].iov_base=hdrbuf;

	iov[1].iov_len=datalen;
	iov[1].iov_base=databuf;

	msg.msg_iovlen=2;
	msg.msg_iov=iov;

	rx_len = recvmsg(sock,&msg,flag);

	if (rx_len <= sizeof(sangoma_api_hdr_t)){
		return -EINVAL;
	}

	rx_len-=sizeof(sangoma_api_hdr_t);

        return rx_len;
}

#ifdef WANPIPE_TDM_API

/*========================================================
 * Execute TDM command
 *
 */
static int sangoma_tdm_cmd_exec(int sock, wanpipe_tdm_api_t *tdm_api)
{
	int err;

	err = ioctl(sock,SIOC_WANPIPE_TDM_API,tdm_api);
	if (err < 0){
		char tmp[50];
		sprintf(tmp,"TDM API: CMD: %i\n",tdm_api->cmd);
		perror(tmp);
		return -1;
	}

	return err;
}

/*========================================================
 * Get Full TDM API configuration per channel
 *
 */
int sangoma_get_full_cfg(int sock, wanpipe_tdm_api_t *tdm_api)
{
	int err;

	tdm_api->cmd = SIOC_WP_TDM_GET_FULL_CFG;

	err=sangoma_tdm_cmd_exec(sock,tdm_api);
	if (err){
		return err;
	}

	printf("TDM API CFG:\n");
	printf("\thw_tdm_coding:\t%d\n",tdm_api->hw_tdm_coding);
	printf("\thw_mtu_mru:\t%d\n",tdm_api->hw_mtu_mru);
	printf("\thw_period:\t%d\n",tdm_api->hw_period);
	printf("\ttdm_codec:\t%d\n",tdm_api->tdm_codec);
	printf("\tpower_level:\t%d\n",tdm_api->power_level);
	printf("\trx_disable:\t%d\n",tdm_api->rx_disable);
	printf("\ttx_disable:\t%d\n",tdm_api->tx_disable);
	printf("\tusr_mtu_mru:\t%d\n",tdm_api->usr_mtu_mru);

	return 0;
}

/*========================================================
 * SET Codec on a particular Channel.
 * 
 * Available codecs are defined in 
 * /usr/src/linux/include/linux/wanpipe_cfg.h
 * 
 * enum wan_codec_format {
 *  	WP_NONE,
 *	WP_SLINEAR
 * }
 *
 */
int sangoma_tdm_set_codec(int sock, wanpipe_tdm_api_t *tdm_api, int codec)
{
	int err;

	tdm_api->cmd = SIOC_WP_TDM_SET_CODEC;
	tdm_api->tdm_codec = codec;

	err=sangoma_tdm_cmd_exec(sock,tdm_api);

	return err;
}

/*========================================================
 * GET Codec from a particular Channel.
 * 
 * Available codecs are defined in 
 * /usr/src/linux/include/linux/wanpipe_cfg.h
 * 
 * enum wan_codec_format {
 *  	WP_NONE,
 *	WP_SLINEAR
 * }
 *
 */
int sangoma_tdm_get_codec(int sock, wanpipe_tdm_api_t *tdm_api)
{
	int err;

	tdm_api->cmd = SIOC_WP_TDM_GET_CODEC;

	err=sangoma_tdm_cmd_exec(sock,tdm_api);
	if (err){
		return err;
	}

	return tdm_api->tdm_codec;	
}


/*========================================================
 * SET Rx/Tx Hardware Period in milliseconds.
 * 
 * Available options are:
 *  10,20,30,40,50 ms      
 *
 */
int sangoma_tdm_set_hw_period(int sock, wanpipe_tdm_api_t *tdm_api, int period)
{
	int err;

	tdm_api->cmd = SIOC_WP_TDM_SET_HW_PERIOD;
	tdm_api->hw_period = period;

	err=sangoma_tdm_cmd_exec(sock,tdm_api);

	return err;
}

/*========================================================
 * GET Rx/Tx Hardware Period in milliseconds.
 * 
 * Available options are:
 *  10,20,30,40,50 ms      
 *
 */
int sangoma_tdm_get_hw_period(int sock, wanpipe_tdm_api_t *tdm_api)
{
	int err;

	tdm_api->cmd = SIOC_WP_TDM_GET_HW_PERIOD;

	err=sangoma_tdm_cmd_exec(sock,tdm_api);
	if (err){
		return err;
	}

	return tdm_api->hw_period;
}

/*========================================================
 * GET Current User MTU/MRU values in bytes.
 * 
 * The USER MTU/MRU values will change each time a PERIOD
 * or CODEC is adjusted.
 */
int sangoma_tdm_get_usr_mtu_mru(int sock, wanpipe_tdm_api_t *tdm_api)
{
	int err;

	tdm_api->cmd = SIOC_WP_TDM_GET_USR_MTU_MRU;

	err=sangoma_tdm_cmd_exec(sock,tdm_api);
	if (err){
		return err;
	}

	return tdm_api->usr_mtu_mru;
}

/*========================================================
 * SET TDM Power Level
 * 
 * This option is not implemented yet
 *
 */
int sangoma_tdm_set_power_level(int sock, wanpipe_tdm_api_t *tdm_api, int power)
{
	int err;

	tdm_api->cmd = SIOC_WP_TDM_SET_POWER_LEVEL;
	tdm_api->power_level = power;

	err=sangoma_tdm_cmd_exec(sock,tdm_api);

	return err;
}

/*========================================================
 * GET TDM Power Level
 * 
 * This option is not implemented yet
 *
 */
int sangoma_tdm_get_power_level(int sock, wanpipe_tdm_api_t *tdm_api)
{
	int err;

	tdm_api->cmd = SIOC_WP_TDM_GET_POWER_LEVEL;

	err=sangoma_tdm_cmd_exec(sock,tdm_api);
	if (err){
		return err;
	}

	return tdm_api->power_level;
}

int sangoma_tdm_flush_bufs(int sock, wanpipe_tdm_api_t *tdm_api)
{
	int err = ioctl(sock,SIOC_WANPIPE_SOCK_FLUSH_BUFS,NULL);
	if (err < 0){
		char tmp[50];
		sprintf(tmp,"TDM API: CMD: %i\n",tdm_api->cmd);
		perror(tmp);
		return -1;
	}

	return 0;
}

#endif /* WANPIPE_TDM_API */
