/*********************************************************************************
 * sangoma_mgd.c --  Sangoma Media Gateway Daemon for Sangoma/Wanpipe Cards
 *
 * Copyright 2005, Anthony Minessale II <anthmct@yahoo.com>
 *
 * This program is free software, distributed under the terms of
 * the GNU General Public License
 *
 *********************************************************************************/


#include <string.h>
#include <stdarg.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <stdio.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/time.h>
#include <poll.h>
#include <signal.h>
#include <pthread.h>
#include <netinet/tcp.h>
#include <libsangoma.h>
#include <assert.h>
#include <call_signal.h>

#if 0
#define DOTRACE
#endif

/* enable early media */
#if 1
#define WOOMERA_EARLY_MEDIA 1
#endif

#ifdef DOTRACE
static int tc = 0;
#endif
#define PIDFILE "/var/run/sangoma_mgd.pid"
#define CORE_EVENT_LEN 512
#define WOOMERA_STRLEN 256
#define WOOMERA_ARRAY_LEN 50
#define WOOMERA_BODYLEN 2048
#define WOOMERA_MIN_MEDIA_PORT 9000
#define WOOMERA_MAX_MEDIA_PORT 9899
#define WOOMERA_HARD_TIMEOUT 0
#define WOOMERA_LINE_SEPERATOR "\r\n"
#define WOOMERA_RECORD_SEPERATOR "\r\n\r\n"
#define WOOMERA_DEBUG_PREFIX "[DEBUG] "
#define WOOMERA_DEBUG_LINE "------------------------------------------------------------------------------------------------"
#define STDERR fileno(stderr)
#define MAXPENDING 500
#define MGD_STACK_SIZE 1024 * 512

const char WELCOME_TEXT[] =
"================================================================================\n"
"Sangoma Media Gateway Daemon -- TDM Signal Media Gateway for Sangoma/Wanpipe Cards\n"
"Copyright 2005, 2006, Anthony Minessale II <anthmct@yahoo.com>\n"
"This program is free software, distributed under the terms of\n"
"the GNU General Public License\n"
"================================================================================\n"
"";



typedef enum {
    WFLAG_RUNNING = (1 << 0),
    WFLAG_LISTENING = (1 << 1),
    WFLAG_EVENT = (1 << 2),
    WFLAG_MALLOC = (1 << 3),
    WFLAG_MEDIA_RUNNING = (1 << 4),
    WFLAG_MEDIA_END = (1 << 5),
    WFLAG_MONITOR_RUNNING = (1 << 6),
	WFLAG_HANGUP = (1 << 7),
	WFLAG_ANSWER = (1 << 8),
} WFLAGS;

typedef enum {
    MFLAG_EXISTS = (1 << 0),
    MFLAG_CONTENT = (1 << 1),
} MFLAGS;

typedef enum {
    EVENT_FREE_DATA = 1,
    EVENT_KEEP_DATA = 0
} event_args;

#define woomera_test_flag(p,flag)		({ \
					((p)->flags & (flag)); \
					})

#define woomera_set_flag(p,flag)		do { \
					((p)->flags |= (flag)); \
					} while (0)

#define woomera_clear_flag(p,flag)		do { \
					((p)->flags &= ~(flag)); \
					} while (0)

#define woomera_copy_flags(dest,src,flagz)	do { \
					(dest)->flags &= ~(flagz); \
					(dest)->flags |= ((src)->flags & (flagz)); \
					} while (0)


static struct woomera_interface woomera_holding_dev;
static struct woomera_interface woomera_dead_dev;
static struct woomera_interface woomera_dead_timeout_dev;

struct media_session {
    int span;
    int chan;
    int udp_sock;
    int sangoma_sock;
    char *ip;
    int port;
    time_t started;
    time_t answered;
    pthread_t thread;
    int socket;
    struct woomera_interface *woomera;
    struct sockaddr_in local_addr;
    struct sockaddr_in remote_addr;
    struct hostent remote_hp;
    struct hostent local_hp;
};

struct woomera_message {
    char callid[WOOMERA_STRLEN];
    char command[WOOMERA_STRLEN];
    char command_args[WOOMERA_STRLEN];
    char names[WOOMERA_STRLEN][WOOMERA_ARRAY_LEN];
    char values[WOOMERA_STRLEN][WOOMERA_ARRAY_LEN];
    char body[WOOMERA_BODYLEN];
    uint32_t flags;
    int last;
    struct woomera_message *next;
};

struct woomera_event {
    char *data;
    uint32_t flags;
    struct woomera_event *next;
};

struct woomera_listener {
    struct woomera_interface *woomera;
    struct woomera_listener *next;
};

struct woomera_interface {
    int socket;
    char *raw;
    char *interface;
    time_t timeout;
    struct sockaddr_in addr;
    struct media_session *ms;
    pthread_mutex_t queue_lock;
    pthread_mutex_t ms_lock;
    struct woomera_event *event_queue;
    struct woomera_event *incoming_event_queue;
    pthread_t thread;
    uint32_t flags;
    int debug;
    int call_id;
	char *cause;
    FILE *log;
    int index;
	int span;
	int chan;
	int call_count;
    struct woomera_interface *next;
};

#define CORE_TANK_LEN CORE_MAX_CHAN_PER_SPAN*CORE_MAX_SPANS

static struct woomera_server {
    struct woomera_interface *process_table[CORE_MAX_CHAN_PER_SPAN][CORE_MAX_SPANS];
    struct woomera_interface *holding_tank[CORE_TANK_LEN];
    int holding_tank_index;
    struct woomera_interface master_connection;
    pthread_mutex_t listen_lock;
    pthread_mutex_t ht_lock;
    pthread_mutex_t process_lock;
    pthread_mutex_t media_udp_port_lock;
    call_signal_connection_t mcon;
    struct woomera_listener *listeners;
    char media_ip[WOOMERA_STRLEN];
    int port;
    int next_media_port;
    int debug;
    int panic;
    int thread_count;
    char *logfile_path;
    FILE *log;
    char boost_local_ip[25];
    int boost_local_port;
    char boost_remote_ip[25];
    int boost_remote_port;
    pthread_t monitor_thread;
    char *config_file;
	int max_calls;
	int call_count;
} server;

struct woomera_config {
    FILE *file;
    char *path;
    char category[256];
    char buf[1024];
    int lineno;
};


static uint32_t string_to_release(char *code)
{
	if (code) {
		if (!strcasecmp(code, "CHANUNAVAIL")) {
			return SIGBOOST_RELEASE_CAUSE_NOANSWER;
		}

		if (!strcasecmp(code, "INVALID")) {
			return SIGBOOST_RELEASE_CAUSE_CALLED_NOT_EXIST;
		}

		if (!strcasecmp(code, "CONGESTION")) {
			//return SIGBOOST_RELEASE_CAUSE_BUSY;
			return SIGBOOST_RELEASE_CAUSE_CALLED_NOT_EXIST;
		}

		if (!strcasecmp(code, "BUSY")) {
			return SIGBOOST_RELEASE_CAUSE_BUSY;
		}

		if (!strcasecmp(code, "NOANSWER")) {
			return SIGBOOST_RELEASE_CAUSE_NOANSWER;
		}

		if (!strcasecmp(code, "ANSWER")) {
			return SIGBOOST_RELEASE_CAUSE_NORMAL;
		}

		if (!strcasecmp(code, "CANCEL")) {
			return SIGBOOST_RELEASE_CAUSE_BUSY;
		}

		if (!strcasecmp(code, "UNKNOWN")) {
			return SIGBOOST_RELEASE_CAUSE_UNDEFINED;
		} 
	}
	return SIGBOOST_RELEASE_CAUSE_NORMAL;
}

static void close_socket(int *sp) 
{
    if (*sp > -1) {
		close(*sp);
		*sp = -1;
    }
}




#define ysleep(usec) sched_yield() ; usleep(usec);
#define log_printf(level, fp, fmt, ...) __log_printf(level, fp, __FILE__, __FUNCTION__, __LINE__, fmt, ##__VA_ARGS__)

static void __log_printf(int level, FILE *fp, char *file, const char *func, int line, char *fmt, ...)
{
    char *data;
    int ret = 0;
    va_list ap;

    if (socket < 0) {
		return;
    }

    if (level && level >= server.debug) {
		return;
    }

    va_start(ap, fmt);
#ifdef SOLARIS
    data = (char *) malloc(2048);
    vsnprintf(data, 2048, fmt, ap);
#else
    ret = vasprintf(&data, fmt, ap);
#endif
    va_end(ap);
    if (ret == -1) {
		fprintf(stderr, "Memory Error\n");
    } else {
		char date[80] = "";
		struct tm now;
		time_t epoch;

		if (time(&epoch) && localtime_r(&epoch, &now)) {
			strftime(date, sizeof(date), "%Y-%m-%d %T", &now);
		}
		
		fprintf(fp, "[%d] %s %s:%d %s() %s", getpid(), date, file, line, func, data);
		free(data);
    }
    fflush(fp);
}

static int socket_printf(int socket, char *fmt, ...)
{
    char *data;
    int ret = 0;
    va_list ap;

    if (socket < 0) {
		return -1;
    }
	
    va_start(ap, fmt);
#ifdef SOLARIS
    data = (char *) malloc(2048);
    vsnprintf(data, 2048, fmt, ap);
#else
    ret = vasprintf(&data, fmt, ap);
#endif
    va_end(ap);
    if (ret == -1) {
		fprintf(stderr, "Memory Error\n");
		log_printf(0, server.log, "Crtical ERROR: Memory Error!\n");
    } else {
		int err;
		int len = strlen(data);
		err=send(socket, data, strlen(data), 0);
		if (err != strlen(data)) {
			log_printf(0, server.log, "ERROR: Failed to send data to woomera socket(%i): err=%i  len=%d %s\n",
					   socket,err,len,strerror(errno));
			ret = err;
		}
		free(data);
    }

	return ret;
}

static FILE *safe_fopen(char *path, char *flags)
{
    char buf[512] = "";

    if (readlink(path, buf, sizeof(buf)) > 0) {
		fprintf(stderr, "Symlinks not allowed! [%s] != [%s]\n", buf, path);
		return NULL;
    }
	
    return fopen(path, flags);
	
}

static int get_pid_from_file(char *path)
{
    FILE *tmp;
    int pid;

    if (!(tmp = safe_fopen(path, "r"))) {
		return 0;
    } else {
		fscanf(tmp, "%d", &pid);
		fclose(tmp);
		tmp = NULL;
    }

    return pid;
}



static int woomera_open_file(struct woomera_config *cfg, char *path) 
{
    FILE *f;

    if (!(f = fopen(path, "r"))) {
		log_printf(0, stderr, "Cannot open file %s\n", path);
		return 0;
    }

    memset(cfg, 0, sizeof(*cfg));
    cfg->file = f;
    cfg->path = path;
    return 1;

}


static void woomera_close_file(struct woomera_config *cfg)
{
	
    if (cfg->file) {
		fclose(cfg->file);
    }

    memset(cfg, 0, sizeof(*cfg));
}



static int woomera_next_pair(struct woomera_config *cfg, char **var, char **val)
{
    int ret = 0;
    char *p, *end;

    *var = *val = NULL;
	
    for(;;) {
		cfg->lineno++;

		if (!fgets(cfg->buf, sizeof(cfg->buf), cfg->file)) {
			ret = 0;
			break;
		}
		
		*var = cfg->buf;

		if (**var == '[' && (end = strchr(*var, ']'))) {
			*end = '\0';
			(*var)++;
			strncpy(cfg->category, *var, sizeof(cfg->category) - 1);
			continue;
		}
		
		if (**var == '#' || **var == '\n' || **var == '\r') {
			continue;
		}

		if ((end = strchr(*var, '#'))) {
			*end = '\0';
			end--;
		} else if ((end = strchr(*var, '\n'))) {
			if (*end - 1 == '\r') {
				end--;
			}
			*end = '\0';
		}
	
		p = *var;
		while ((*p == ' ' || *p == '\t') && p != end) {
			*p = '\0';
			p++;
		}
		*var = p;

		if (!(*val = strchr(*var, '='))) {
			ret = -1;
			log_printf(0, server.log, "Invalid syntax on %s: line %d\n", cfg->path, cfg->lineno);
			continue;
		} else {
			p = *val - 1;
			*(*val) = '\0';
			(*val)++;
			if (*(*val) == '>') {
				*(*val) = '\0';
				(*val)++;
			}

			while ((*p == ' ' || *p == '\t') && p != *var) {
				*p = '\0';
				p--;
			}

			p = *val;
			while ((*p == ' ' || *p == '\t') && p != end) {
				*p = '\0';
				p++;
			}
			*val = p;
			ret = 1;
			break;
		}
    }

    return ret;

}



static void woomera_set_raw(struct woomera_interface *woomera, char *raw)
{

    if (woomera->raw) {
		free(woomera->raw);
		woomera->raw = NULL;
    }

    if (raw) {
		woomera->raw = strdup(raw);
    }
}


static void woomera_set_interface(struct woomera_interface *woomera, char *interface)
{

    if (woomera->interface) {
		free(woomera->interface);
		woomera->interface = NULL;
    }

    if (interface) {
		woomera->interface = strdup(interface);
    }

}

static void woomera_set_cause(struct woomera_interface *woomera, char *cause)
{

    if (woomera->cause) {
		free(woomera->cause);
		woomera->cause = NULL;
    }

    if (cause) {
		woomera->cause = strdup(cause);
    }

}

static struct woomera_event *new_woomera_event(void)
{
    struct woomera_event *event = NULL;

    if ((event = malloc(sizeof(*event)))) {
		memset(event, 0, sizeof(*event));
		woomera_set_flag(event, WFLAG_MALLOC);
    }

    return event;
}


static void destroy_woomera_event(struct woomera_event **event, event_args free_data)
{
    if (free_data) {
		free ((*event)->data);
    }
    if (woomera_test_flag((*event), WFLAG_MALLOC)) {
		free (*event);
		*event = NULL;
    }
}

static struct woomera_event *new_woomera_event_printf(struct woomera_event *ebuf, char *fmt, ...)
{
    struct woomera_event *event = NULL;
    int ret = 0;
    va_list ap;

    if (ebuf) {
		event = ebuf;
    } else if (!(event = new_woomera_event())) {
		log_printf(0, server.log, "Memory Error queuing event!\n");
		return NULL;
    }

    va_start(ap, fmt);
#ifdef SOLARIS
    event->data = (char *) malloc(2048);
    vsnprintf(event->data, 2048, fmt, ap);
#else
    ret = vasprintf(&event->data, fmt, ap);
#endif
    va_end(ap);
    if (ret == -1) {
		log_printf(0, server.log, "Memory Error queuing event!\n");
		destroy_woomera_event(&event, EVENT_FREE_DATA);
		return NULL;
    }

    return event;
	
}

static struct woomera_event *woomera_clone_event(struct woomera_event *event)
{
    struct woomera_event *clone;

    if (!(clone = new_woomera_event())) {
		return NULL;
    }

    memcpy(clone, event, sizeof(*event));
    clone->next = NULL;
    clone->data = strdup(event->data);

    return clone;
}

static void enqueue_event(struct woomera_interface *woomera, struct woomera_event *event)
{
    struct woomera_event *ptr, *clone = NULL;
	
    assert(woomera != NULL);
    assert(event != NULL);

	
    if (!(clone = woomera_clone_event(event))) {
		log_printf(0, server.log, "Error Cloning Event\n");
		return;
    }

    pthread_mutex_lock(&woomera->queue_lock);
	
    for (ptr = woomera->event_queue; ptr && ptr->next ; ptr = ptr->next);
	
    if (ptr) {
		ptr->next = clone;
    } else {
		woomera->event_queue = clone;
    }

    pthread_mutex_unlock(&woomera->queue_lock);

    woomera_set_flag(woomera, WFLAG_EVENT);
}

static char *dequeue_event(struct woomera_interface *woomera) 
{
    struct woomera_event *event;
    char *data = NULL;

    if (!woomera) {
		return NULL;
    }
	
    pthread_mutex_lock(&woomera->queue_lock);
    if (woomera->event_queue) {
		event = woomera->event_queue;
		woomera->event_queue = event->next;
		data = event->data;
		destroy_woomera_event(&event, EVENT_KEEP_DATA);
    }
    pthread_mutex_unlock(&woomera->queue_lock);

    return data;
}


static int enqueue_event_on_listeners(struct woomera_event *event) 
{
    struct woomera_listener *ptr;
    int x = 0;

    assert(event != NULL);

    pthread_mutex_lock(&server.listen_lock);
    for (ptr = server.listeners ; ptr ; ptr = ptr->next) {
		enqueue_event(ptr->woomera, event);
		x++;
    }
    pthread_mutex_unlock(&server.listen_lock);
	
    return x;
}


static void del_listener(struct woomera_interface *woomera) 
{
    struct woomera_listener *ptr, *last = NULL;

    pthread_mutex_lock(&server.listen_lock);
    for (ptr = server.listeners ; ptr ; ptr = ptr->next) {
		if (ptr->woomera == woomera) {
			if (last) {
				last->next = ptr->next;
			} else {
				server.listeners = ptr->next;
			}
			free(ptr);
			break;
		}
		last = ptr;
    }
    pthread_mutex_unlock(&server.listen_lock);
}

static void add_listener(struct woomera_interface *woomera) 
{
    struct woomera_listener *new;

    pthread_mutex_lock(&server.listen_lock);
	
    if ((new = malloc(sizeof(*new)))) {
		memset(new, 0, sizeof(*new));
		new->woomera = woomera;
		new->next = server.listeners;
		server.listeners = new;
    } else {
		log_printf(0, server.log, "Memory Error adding listener!\n");
    }

    pthread_mutex_unlock(&server.listen_lock);
}

/* disable nagle's algorythm */
static void no_nagle(int socket)
{
    int flag = 1;
    setsockopt(socket, IPPROTO_TCP, TCP_NODELAY, (char *) &flag, sizeof(int));
}

static struct woomera_interface *new_woomera_interface(int socket, struct sockaddr_in *sock_addr, int len) 
{
    struct woomera_interface *woomera = NULL;
	
    if (socket < 0) {
		log_printf(0, server.log, "Critical: Invalid Socket on new interface!\n");
		return NULL;
    }


    if ((woomera = malloc(sizeof(*woomera)))) {
		memset(woomera, 0, sizeof(*woomera));
		server.call_count++;
		if (socket >= 0) {
			no_nagle(socket);
			woomera->socket = socket;
		}
		if (sock_addr && len) {
			memcpy(&woomera->addr, sock_addr, len);
		}
    
		woomera->chan = -1;
		woomera->span = -1;
		woomera->log = server.log;
		woomera->debug = server.debug;
		woomera->call_id = 1;
		woomera->event_queue = NULL;
	}

	return woomera;

}

static char *woomera_message_header(struct woomera_message *wmsg, char *key) 
{
    int x = 0;
    char *value = NULL;

    for (x = 0 ; x < wmsg->last ; x++) {
		if (!strcasecmp(wmsg->names[x], key)) {
			value = wmsg->values[x];
			break;
		}
    }

    return value;
}


static int waitfor_socket(int fd, int timeout, int flags)
{
    struct pollfd pfds[1];
    int res;
    int errflags = (POLLERR | POLLHUP | POLLNVAL);


    memset(&pfds[0], 0, sizeof(pfds[0]));
    pfds[0].fd = fd;
    pfds[0].events = flags | errflags;
    res = poll(pfds, 1, timeout);

    if (res > 0) {
	if(pfds[0].revents & POLLIN) {
		res = 1;
	} else if ((pfds[0].revents & errflags)) {
		res = -1;
    	} else {
		/* No event found what to do */
	}
    }

    return res;
}

static int waitfor_2sockets(int fda, int fdb, int timeout) 
{
    struct pollfd pfds[2];
    int res = 0;
    int errflags = (POLLERR | POLLHUP | POLLNVAL);

    if (fda < 0 || fdb < 0) {
		return -1;
    }

    memset(pfds, 0, sizeof(pfds));
    pfds[0].fd = fda;
    pfds[1].fd = fdb;
    pfds[0].events = POLLIN | errflags;
    pfds[1].events = POLLIN | errflags;
    if ((res = poll(pfds, 2, timeout)) > 0) {
		res = 1;
		if ((pfds[0].revents & errflags) || (pfds[1].revents & errflags)) {
			res = -1;
		} else { 
			if ((pfds[0].revents & POLLIN)) {
				res += fda;
			}
			if ((pfds[1].revents & POLLIN)) {
				res += fdb;
			}
		}
		if (res == 1) {
			/* No event found what to do */
		}
    }
	
    return res;
}

static struct media_session *media_session_new(struct woomera_interface *woomera)
{
    struct media_session *ms = NULL;
    int x;
    char *p;
    int span,chan;
    
    sangoma_interface_toi(woomera->interface, &span, &chan);
    span--;
    chan--;
    
  
    
    log_printf(2, server.log,"Starting new MEDIA session [%s]\n",
						woomera->interface);
    
    if (span < 0 || chan < 0) {
    	log_printf(0, server.log, "ERROR: Starting MEDIA on Invalid: Span=%i Chan=%i!\n",
			span+1,chan+1);	
		return NULL;	
    }  
						
    if (server.process_table[span][chan]) {
    
    	if (server.process_table[span][chan] == &woomera_dead_timeout_dev) {
		log_printf(2, server.log, "Starting MEDIA over TIMEOUT Device Span=%i Chan=%i [w%dg%d]!\n",
			span+1,chan+1,woomera->interface);
		server.process_table[span][chan]=NULL;
	} else {
    
		log_printf(0, server.log, "Critical Error: Span=%i Chan=%i Overrun (O=%p N=%p: H=%p D=%p DT=%p)!\n",
			span+1,chan+1,
			server.process_table[span][chan],
			woomera,
			&woomera_holding_dev,
			&woomera_dead_dev,
			&woomera_dead_timeout_dev);
			
		return NULL;
    	}
    }

        if ((ms = malloc(sizeof *ms))) {
		memset(ms, 0, sizeof(*ms));
		for(x = 0; x < strlen(woomera->raw) ; x++) {
			if (woomera->raw[x] == '/') {
				break;
			}
		}
		ms->ip = strndup(woomera->raw, x);
		time(&ms->started);
		p = woomera->raw + (x+1);
		ms->port = atoi(p);
		time(&ms->started);
		woomera->ms = ms;
		ms->woomera = woomera;
		sangoma_interface_toi(woomera->interface, &ms->span, &ms->chan);
		ms->span--;
		ms->chan--;
		
		server.process_table[ms->span][ms->chan] = woomera;
    }

    return ms;
}

static void media_session_free(struct media_session *ms) 
{
    if (ms->ip) {
		free(ms->ip);
    }

    ms->woomera = NULL;

    free(ms);
}


static int create_udp_socket(struct media_session *ms, char *local_ip, int local_port, char *ip, int port)
{
    int rc;
    struct hostent *result, *local_result;
    char buf[512], local_buf[512];
    int err = 0;

    memset(&ms->remote_hp, 0, sizeof(ms->remote_hp));
    memset(&ms->local_hp, 0, sizeof(ms->local_hp));
    if ((ms->socket = socket(AF_INET, SOCK_DGRAM, 0))) {
		gethostbyname_r(ip, &ms->remote_hp, buf, sizeof(buf), &result, &err);
		gethostbyname_r(local_ip, &ms->local_hp, local_buf, sizeof(local_buf), &local_result, &err);
		if (result && local_result) {
			ms->remote_addr.sin_family = ms->remote_hp.h_addrtype;
			memcpy((char *) &ms->remote_addr.sin_addr.s_addr, ms->remote_hp.h_addr_list[0], ms->remote_hp.h_length);
			ms->remote_addr.sin_port = htons(port);

			ms->local_addr.sin_family = ms->local_hp.h_addrtype;
			memcpy((char *) &ms->local_addr.sin_addr.s_addr, ms->local_hp.h_addr_list[0], ms->local_hp.h_length);
			ms->local_addr.sin_port = htons(local_port);

			if ((rc = bind(ms->socket, (struct sockaddr *) &ms->local_addr, sizeof(ms->local_addr))) < 0) {
				close(ms->socket);
				ms->socket = -1;
			}
		}
    }

    return ms->socket;
}

static int next_media_port(void)
{
    int port;
    
    pthread_mutex_lock(&server.media_udp_port_lock);
    port = ++server.next_media_port;
    if (port > WOOMERA_MAX_MEDIA_PORT) {
    		server.next_media_port = WOOMERA_MIN_MEDIA_PORT;
		port = WOOMERA_MIN_MEDIA_PORT;
    }
    pthread_mutex_unlock(&server.media_udp_port_lock);
    
    return port;
}

static void *media_thread_run(void *obj)
{
    struct media_session *ms = obj;
    struct woomera_interface *woomera = ms->woomera;
    int circuit_frame_len = 0;
    int udp_frame_len = 0;
    int sangoma_frame_len = 160;
    int local_port, x = 0, errs = 0, res = 0, packet_len = 0;
    struct woomera_event wevent;
    wanpipe_tdm_api_t tdm_api;

    if (woomera_test_flag(woomera, WFLAG_MEDIA_END) || !woomera->interface) {
	log_printf(1, server.log, "MEDIA session for [%s] Cancelled! (ptr=%p)\n", woomera->interface,woomera);
	/* In this case the call will be closed via woomera_thread_run
	 * function. And the process table will be cleard there */
	woomera_set_flag(woomera, WFLAG_MEDIA_END);
	return NULL;
    }


    log_printf(1, server.log, "MEDIA session for [%s] started (ptr=%p)\n", woomera->interface,woomera);
    woomera_set_flag(woomera, WFLAG_MEDIA_RUNNING);

    for(x = 0; x < 1000 ; x++) {
		local_port = next_media_port();
		if ((ms->udp_sock = create_udp_socket(ms, server.media_ip, local_port, ms->ip, ms->port)) > -1) {
			break;
		}
    }

    if (ms->udp_sock < 0) {
		log_printf(0, server.log, "UDP Socket Error (%s) [%s] LocalPort=%d\n", 
				strerror(errno), woomera->interface, local_port);
#if 0
		if (errno == EADDRINUSE) {
			/* Do not try to stop the call if its already in use,
			 * Indicate to this channel that is already hungup */
			woomera_set_flag(woomera, WFLAG_HANGUP);
		}
#endif
		errs++;
    } else {
		if ((ms->sangoma_sock = sangoma_create_socket_by_name(woomera->interface, NULL)) < 0) {
			log_printf(0, server.log, "WANPIPE Socket Error (%s) if=[%s]  [w%ig%i]\n", 
				strerror(errno), woomera->interface, woomera->span+1, woomera->chan+1);
			errs++;
		} else {
			if (sangoma_tdm_set_codec(ms->sangoma_sock, &tdm_api, WP_SLINEAR) < 0 ) {
				errs++;	
			}

			if (sangoma_tdm_flush_bufs(ms->sangoma_sock, &tdm_api)) {					
				errs++;
			}
#if 0
			/* FIXME: try to use the mtu .... */
			mtu_mru = sangoma_tdm_get_usr_mtu_mru(ms->sangoma_sock,&tdm_api);
#endif
		}
    }
	
    if (!errs) {
		unsigned char udp_frame[1024], circuit_frame[1024], *udp_frame_p, *circuit_frame_p, *frame_p;
		sangoma_api_hdr_t hdrframe;
#ifdef DOTRACE
		int fdin, fdout;
		char path_in[512], path_out[512];
#endif
		new_woomera_event_printf(&wevent, 
								 "EVENT MEDIA %s AUDIO%s"
								 "Raw-Audio: %s/%d%s"
								 "Call-ID: %s%s"
								 "Raw-Format: PCM-16%s"
								 ,
								 woomera->interface,
								 WOOMERA_LINE_SEPERATOR,
								 server.media_ip,
								 local_port,
								 WOOMERA_LINE_SEPERATOR,
								 woomera->interface,
								 WOOMERA_LINE_SEPERATOR,
								 WOOMERA_RECORD_SEPERATOR
								 );


		enqueue_event(woomera, &wevent);

#ifdef DOTRACE
		sprintf(path_in, "/tmp/debug-in.%d.raw", tc);
		sprintf(path_out, "/tmp/debug-out.%d.raw", tc++);
		fdin = open(path_in, O_WRONLY | O_CREAT, O_TRUNC, 0600);
		fdout = open(path_out, O_WRONLY | O_CREAT, O_TRUNC, 0600);
#endif

		udp_frame_p = udp_frame;
		circuit_frame_p = circuit_frame;
		frame_p = udp_frame;

		udp_frame_len = circuit_frame_len = 0;
		while ( woomera_test_flag(&server.master_connection, WFLAG_RUNNING) && !woomera_test_flag(woomera, WFLAG_MEDIA_END) && (res = waitfor_2sockets(ms->sangoma_sock, ms->udp_sock, 100)) > -1) {
			int fromlen = sizeof(struct sockaddr_in);
			int circuit_event;
			int woomera_event;
			
			if (res == 0) {
				/* NENAD Timeout thus just continue */
				continue;
			}
			
			/* Decrement 1 because the result of waitfor_2sockets 
			 * is incremented from 1 */
			res--;
			
			circuit_event = (res == ms->sangoma_sock || res == (ms->sangoma_sock + ms->udp_sock)) ? 1 : 0;
			woomera_event= (res == ms->udp_sock || res == (ms->sangoma_sock + ms->udp_sock)) ? 1 : 0;
			
			
			if (woomera_event) {
				if ((packet_len = recvfrom(ms->udp_sock, udp_frame_p, sizeof(udp_frame), 
				     MSG_DONTWAIT, (struct sockaddr *) &ms->local_addr, &fromlen)) < 1) {
					log_printf(0, server.log, "Bad Data from Media UDP\n");
					break;
				}
				udp_frame_len += packet_len;
				udp_frame_p += udp_frame_len;
			}
			
			if (circuit_event) {
				res = sangoma_readmsg_socket(ms->sangoma_sock, &hdrframe, sizeof(hdrframe), circuit_frame_p, sangoma_frame_len,	 0);
				if (res < 0) {
					log_printf(0, server.log, "Bad Data from Sangoma %d\n", errno);
					break;
				}
				circuit_frame_len += res;
				circuit_frame_p += circuit_frame_len;
			}

			
			
			while (udp_frame_len) {
				int len = (udp_frame_len >= sangoma_frame_len) ? sangoma_frame_len : udp_frame_len;
				sangoma_sendmsg_socket(ms->sangoma_sock, &hdrframe, sizeof(hdrframe), frame_p, len, 0);
#ifdef DOTRACE
				write(fdin, frame_p , len);
#endif
				frame_p += len;
				memmove(udp_frame, frame_p, sizeof(udp_frame) - len);
				frame_p = udp_frame_p = udp_frame;
				udp_frame_len -= len;
			}
			
			if (circuit_frame_len >= packet_len) {
				//log_printf(0, server.log, "XXX write %d\n", circuit_frame_len);
				sendto(ms->udp_sock, circuit_frame, circuit_frame_len, 0, (struct sockaddr *) &ms->remote_addr, sizeof(ms->remote_addr));
#ifdef DOTRACE
				write(fdout, circuit_frame , circuit_frame_len);
#endif
				circuit_frame_len = 0;
				circuit_frame_p = circuit_frame;
			}
	

		}
#ifdef DOTRACE
		close (fdin);
		close (fdout);
#endif
    }

    new_woomera_event_printf(&wevent, 
							 "EVENT HANGUP %s%s"
							 "Start-Time: %ld%s"
							 "End-Time: %ld%s"
							 "Answer-Time: %ld%s"
							 "Call-ID: %s%s"
							 ,
							 woomera->interface,
							 WOOMERA_LINE_SEPERATOR,

							 time(&ms->started),
							 WOOMERA_LINE_SEPERATOR,

							 time(NULL),
							 WOOMERA_LINE_SEPERATOR,

							 time(&ms->answered),
							 WOOMERA_LINE_SEPERATOR,

							 woomera->interface,
							 WOOMERA_RECORD_SEPERATOR
							 );

    enqueue_event(woomera, &wevent);

    close_socket(&ms->udp_sock);
    close_socket(&ms->sangoma_sock);
    
    pthread_mutex_lock(&server.process_lock);
    server.process_table[ms->span][ms->chan] = NULL;
    pthread_mutex_unlock(&server.process_lock);
    
    log_printf(1, server.log, "MEDIA session for [%s] ended [%d][%d] (ptr=%p)\n", 
		    woomera->interface, res, udp_frame_len,woomera);
    
    woomera_clear_flag(woomera, WFLAG_MEDIA_RUNNING);
    woomera_set_flag(woomera, WFLAG_MEDIA_END);

    if (ms && !woomera_test_flag(woomera, WFLAG_HANGUP)) {
	call_signal_event_t event;
	woomera_set_flag(woomera, WFLAG_HANGUP);
		
	call_signal_event_init(&event, SIGBOOST_EVENT_CALL_STOPPED, ms->chan, ms->span);
	event.release_cause = string_to_release(woomera->cause);
	if (call_signal_connection_write(&server.mcon, &event) <= 0) {
		log_printf(0, server.log, "Critical System Error: Failed to tx on ISUP socket: %s\n", 
						strerror(errno));
	}
	log_printf(2, woomera->log, "Sent (From Media) SIGBOOST_EVENT_CALL_STOPPED %d/%d [%s]\n", 
			ms->span, ms->chan,woomera->interface);
    }

    media_session_free(ms);
    return NULL;
}

/* This function must be called with process_lock 
 * because it modifies shared process_table */

static int launch_media_thread(struct woomera_interface *woomera) 
{
    pthread_attr_t attr;
    int result = -1;
    struct media_session *ms;

    if ((ms = media_session_new(woomera))) {
		result = pthread_attr_init(&attr);
		pthread_attr_setschedpolicy(&attr, SCHED_RR);
		pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_DETACHED);
		pthread_attr_setstacksize(&attr, MGD_STACK_SIZE);
		result = pthread_create(&ms->thread, &attr, media_thread_run, ms);
		result = pthread_attr_destroy(&attr);
		result = 0;
    } else {
		log_printf(0, server.log, "Memory Error\n");
    }
    
    return result;

}

#if 1

static int launch_media_thread_hold_check(struct woomera_interface *woomera) 
{
	int span,chan,err=-1;
	sscanf(woomera->interface, "w%dg%d",  &span, &chan);
	span--;
	chan--;
	
	if (span < 0 || chan < 0 ) {
		log_printf(2, server.log,"ERROR: Incoming Call without SPAN=%i CHAN=%i [%s]\n",
			span+1,chan+1,woomera->interface);
		return -1;
	}
	
	pthread_mutex_lock(&server.process_lock);
	
	/* Check for holding device of incoming call.
	 * If we do not launch media, the woomera thread
	 * will end the call. */
	if (server.process_table[span][chan] != &woomera_holding_dev) {
	
		log_printf(2, server.log,"ERROR: Incoming Call without HOLDINV DEV [%s]\n",
			woomera->interface);
		
		if (server.process_table[span][chan] == &woomera_dead_dev){
			call_signal_event_t event;
			
			server.process_table[span][chan]=NULL;
			pthread_mutex_unlock(&server.process_lock);
			
			log_printf(2, server.log,"ERROR: Incoming Call FOUND DEAD DEV: Clearing SIGBOOST_EVENT_CALL_STOPPED [%s] Span=%i Chan=%i\n",
			woomera->interface,span+1,chan+1);
				
			woomera_set_flag(woomera, WFLAG_HANGUP);
		 	
			call_signal_event_init(&event, SIGBOOST_EVENT_CALL_STOPPED, chan, span);
			event.release_cause = SIGBOOST_RELEASE_CAUSE_NORMAL;
			if (call_signal_connection_write(&server.mcon, &event) <= 0) {
				log_printf(0, server.log, "Critical System Error: Failed to tx on ISUP socket: %s\n", 
						strerror(errno));
			}
			
			woomera_set_flag(woomera, WFLAG_HANGUP);
			woomera_set_flag(woomera, WFLAG_MEDIA_END);
			
			return -1;
			
		} else if (server.process_table[span][chan] == &woomera_dead_timeout_dev) {
			
			log_printf(2, server.log,"ERROR: Incoming Call FOUND TIMEOUT DEAD DEV: Clearing SIGBOOST_EVENT_CALL_STOPPED [%s]\n",
			woomera->interface);
			
			server.process_table[span][chan]=NULL;
			pthread_mutex_unlock(&server.process_lock);
			
			woomera_set_flag(woomera, WFLAG_HANGUP);
			woomera_set_flag(woomera, WFLAG_MEDIA_END);
		
			return -1;	
		} 


		log_printf(2, server.log,"ERROR: Incoming Call Critical Overrun Attempt: [%s] (O=%p N=%p)\n",
				woomera->interface,server.process_table[span][chan],woomera);

		/* In this case, the call is already used by another device
		 * meaning that the call hung up on us, and we are too late */
		woomera_set_flag(woomera, WFLAG_HANGUP);
		woomera_set_flag(woomera, WFLAG_MEDIA_END);
		
		pthread_mutex_unlock(&server.process_lock);

		return -1;
				
	}else{
		/* Clear the holding device from chan span
		 * and launch the new media thread */
		server.process_table[span][chan] = NULL;
		
		/* Note that process_lock has already been taken
		 * around this function so do not call process_lock
		 * here again */
		err=launch_media_thread(woomera);
	}
	
	pthread_mutex_unlock(&server.process_lock);
	
	return err;

}

#endif


static int woomera_message_parse(struct woomera_interface *woomera, struct woomera_message *wmsg, int timeout) 
{
    char *cur, *cr, *next = NULL, *eor = NULL;
    char buf[2048];
    int res = 0, bytes = 0, sanity = 0;
    struct timeval started, ended;
    int elapsed, loops = 0;
    int failto = 0;
    int packet = 0;

    memset(wmsg, 0, sizeof(*wmsg));

    if (woomera->socket < 0 ) {
   	 log_printf(2, woomera->log, WOOMERA_DEBUG_PREFIX "%s Invalid Socket! %d\n", 
			 woomera->interface,woomera->socket);
		return -1;
    }

	if (woomera_test_flag(woomera, WFLAG_MEDIA_END) || woomera_test_flag(woomera, WFLAG_HANGUP)) {
	log_printf(2, woomera->log, WOOMERA_DEBUG_PREFIX "%s MEDIA END or HANGUP !\n", 
			woomera->interface);
		return -1;
	}

    gettimeofday(&started, NULL);
    memset(buf, 0, sizeof(buf));

    if (timeout < 0) {
		timeout = abs(timeout);
		failto = 1;
    } else if (timeout == 0) {
		timeout = -1;
    }
	

    while (!(eor = strstr(buf, WOOMERA_RECORD_SEPERATOR))) {
		if (sanity > 1000) {
			log_printf(1, woomera->log, WOOMERA_DEBUG_PREFIX "%s Failed Sanity Check!\n[%s]\n\n", woomera->interface, buf);
			return -1;
		}

		if ((res = waitfor_socket(woomera->socket, 1000, POLLERR | POLLIN) > 0)) {
			res = recv(woomera->socket, buf, sizeof(buf), MSG_PEEK);

			if (res > 1) {
				packet++;
			}
			if (!strncmp(buf, WOOMERA_LINE_SEPERATOR, 2)) {
				res = read(woomera->socket, buf, 2);
				return 0;
			}
			if (res == 0) {
				sanity++;
				/* Looks Like it's time to go! */
				if (woomera_test_flag(woomera, WFLAG_MEDIA_END) || woomera_test_flag(woomera, WFLAG_HANGUP)) {
				log_printf(2, woomera->log, WOOMERA_DEBUG_PREFIX "%s MEDIA END or HANGUP \n", woomera->interface);
					return -1;
				}
				ysleep(1000);
				continue;
			} else if (res < 0) {
				log_printf(0, woomera->log, WOOMERA_DEBUG_PREFIX "%s error during packet retry #%d\n", woomera->interface, loops);
				return res;
			} else if (loops) {
				ysleep(100000);
			}
		}
		
		gettimeofday(&ended, NULL);
		elapsed = (((ended.tv_sec * 1000) + ended.tv_usec / 1000) - ((started.tv_sec * 1000) + started.tv_usec / 1000));

		if (res < 0) {
			log_printf(0, woomera->log, WOOMERA_DEBUG_PREFIX "%s Bad RECV\n", woomera->interface);
			return res;
		} else if (res == 0) {
			sanity++;
			/* Looks Like it's time to go! */
			if (woomera_test_flag(woomera, WFLAG_MEDIA_END) || woomera_test_flag(woomera, WFLAG_HANGUP)) {
			log_printf(2, woomera->log, WOOMERA_DEBUG_PREFIX "%s MEDIA END or HANGUP \n", woomera->interface);
				return -1;
			}
			ysleep(1000);
			continue;
		}

		if (packet && loops > 150) {
			log_printf(1, woomera->log, WOOMERA_DEBUG_PREFIX "%s Timeout waiting for packet.\n[%s]\n", woomera->interface, buf);
			return -1;
		}

		if (timeout > 0 && (elapsed > timeout)) {
			log_printf(1, woomera->log, WOOMERA_DEBUG_PREFIX "%s Timeout [%d] reached\n", woomera->interface, timeout);
			return failto ? -1 : 0;
		}

		if (woomera_test_flag(woomera, WFLAG_EVENT)) {
			/* BRB! we have an Event to deliver....*/
			return 0;
		}

		/* what're we still doing here? */
		if (!(woomera_test_flag(&server.master_connection, WFLAG_RUNNING) && woomera_test_flag(woomera, WFLAG_RUNNING))) {
		log_printf(2, woomera->log, WOOMERA_DEBUG_PREFIX "%s MASTER RUNNING or RUNNING!\n", woomera->interface);
			return -1;
		}
		loops++;
    }
    *eor = '\0';
    bytes = strlen(buf) + 4;
	
    memset(buf, 0, sizeof(buf));
    res = read(woomera->socket, buf, bytes);
    next = buf;

    if (woomera->debug > 1) {
		log_printf(2, woomera->log, WOOMERA_DEBUG_PREFIX "%s Receive Message:\n%s\n%s", woomera->interface, WOOMERA_DEBUG_LINE, buf);
    }
	
    while ((cur = next)) {


		if ((cr = strstr(cur, WOOMERA_LINE_SEPERATOR))) {
			*cr = '\0';
			next = cr + (sizeof(WOOMERA_LINE_SEPERATOR) - 1);
			if (!strcmp(next, WOOMERA_RECORD_SEPERATOR)) {
				break;
			}
		} 
		if (!cur || !*cur) {
			break;
		}

		if (!wmsg->last) {
			char *cmd, *id, *args;
			woomera_set_flag(wmsg, MFLAG_EXISTS);
			cmd = cur;

			if ((id = strchr(cmd, ' '))) {
				*id = '\0';
				id++;
				if ((args = strchr(id, ' '))) {
					*args = '\0';
					args++;
					strncpy(wmsg->command_args, args, sizeof(wmsg->command_args)-1);
				}
				strncpy(wmsg->callid, id, sizeof(wmsg->callid)-1);
			}

			strncpy(wmsg->command, cmd, sizeof(wmsg->command)-1);
		} else {
			char *name, *val;
			name = cur;

			if ((val = strchr(name, ':'))) {
				*val = '\0';
				val++;
				while (*val == ' ') {
					*val = '\0';
					val++;
				}
				strncpy(wmsg->values[wmsg->last-1], val, WOOMERA_STRLEN);
			}
			strncpy(wmsg->names[wmsg->last-1], name, WOOMERA_STRLEN);
			if (name && val && !strcasecmp(name, "content-type")) {
				woomera_set_flag(wmsg, MFLAG_CONTENT);
				bytes = atoi(val);
			}

		}
		wmsg->last++;
    }

    wmsg->last--;

    if (bytes && woomera_test_flag(wmsg, MFLAG_CONTENT)) {
		read(woomera->socket, wmsg->body, (bytes > sizeof(wmsg->body)) ? sizeof(wmsg->body) : bytes);
    }

    return woomera_test_flag(wmsg, MFLAG_EXISTS);

}

static struct woomera_interface *pull_from_holding_tank(int index)
{
    struct woomera_interface *woomera = NULL;

    pthread_mutex_lock(&server.ht_lock);
    if (server.holding_tank[index]) {
		woomera = server.holding_tank[index];
		server.holding_tank[index] = NULL;
    
    		woomera->timeout = 0;
    		woomera->index = 0;
    }
    pthread_mutex_unlock(&server.ht_lock);

    return woomera;
}

static int add_to_holding_tank(struct woomera_interface *woomera)
{
    pthread_mutex_lock(&server.ht_lock);
    
    /* NENAD Changed to pre increment */
    int next = ++server.holding_tank_index;
    if (server.holding_tank_index >= CORE_TANK_LEN) {
	next = server.holding_tank_index = 1;
    } 
    
    if (next == 0) {
        log_printf(0, server.log, "\nCritical Error on TANK INDEX == 0\n");
    }
    if (server.holding_tank[next]) {
    	log_printf(0, server.log, "\nCritical Error on TANK INDEX %d Overrun\n",
		next);
    }
    
    server.holding_tank[next] = woomera;
    woomera->timeout = time(NULL) + 30;
	
    pthread_mutex_unlock(&server.ht_lock);
    return next;
}




static void interpret_command(struct woomera_interface *woomera, struct woomera_message *wmsg)
{
    int answer = 0, media = 0;
	
    if (!strcasecmp(wmsg->command, "bye") || !strcasecmp(wmsg->command, "quit")) {
		char *cause = woomera_message_header(wmsg, "cause");
		
		if (cause) {
			log_printf(2, woomera->log, "Bye Cause Received: [%s]\n", cause);
		}

		log_printf(2, woomera->log, "WOOMERA CMD: Bye Received: [%s]\n", woomera->interface);
		
		woomera_clear_flag(woomera, WFLAG_RUNNING);
		socket_printf(woomera->socket, "200 Connection closed%s", WOOMERA_RECORD_SEPERATOR);

    } else if (!strcasecmp(wmsg->command, "listen")) {
		if (woomera_test_flag(woomera, WFLAG_LISTENING)) {
			socket_printf(woomera->socket, "405 Listener already started%s", WOOMERA_RECORD_SEPERATOR);
		} else {
			char *event_string;

			woomera_set_flag(woomera, WFLAG_LISTENING);
			add_listener(woomera);
			socket_printf(woomera->socket, "200 Listener enabled%s", WOOMERA_RECORD_SEPERATOR);
			if ((event_string = dequeue_event(&server.master_connection))) {
				socket_printf(woomera->socket, "%s", event_string);
				free(event_string);
				event_string = NULL;
			} 
		}
    } else if (!strcasecmp(wmsg->command, "dtmf")) {
		
		log_printf(2, woomera->log, "WOOMERA CMD: DTMF Received: [%s]  Digit %s Body %s\n", 	
			woomera->interface, wmsg->command_args, wmsg->body);
		woomera_message_header(wmsg, woomera->interface);
		
		socket_printf(woomera->socket, "200 DTMF OK%s", WOOMERA_RECORD_SEPERATOR);
		
    } else if (!strcasecmp(wmsg->command, "hangup")) {
		int chan = -1, span = -1;
		char *cause = woomera_message_header(wmsg, "cause");

		woomera_set_cause(woomera, cause);
		if (woomera->ms) {
			pthread_mutex_lock(&woomera->queue_lock);
			chan = woomera->ms->chan;
			span = woomera->ms->span;
			shutdown(woomera->ms->sangoma_sock, SHUT_RDWR);
			shutdown(woomera->ms->udp_sock, SHUT_RDWR);
			pthread_mutex_unlock(&woomera->queue_lock);
			log_printf(2, woomera->log, "WOOMERA CMD: Hangup Received: [%s] MEDIA EXIST\n", 	woomera->interface);
		} else if (woomera->chan > -1 && woomera->span > -1) {
			chan = woomera->chan;
			span = woomera->span;
			log_printf(2, woomera->log, "WOOMERA CMD: Hangup Received: [%s] NO MEDIA CHAN/SPAN\n", 	woomera->interface);
		} else {
			sscanf(wmsg->callid, "w%dg%d",  &span, &chan);
			span--;
			chan--;
			log_printf(2, woomera->log, "WOOMERA CMD: Hangup Received: [%s] NO MEDIA AND NO CHAN/SPAN\n", 	woomera->interface);
		}

		if (chan > -1 && span > -1) {
			woomera_set_flag(woomera, WFLAG_MEDIA_END);
			
			//Only do a hangup after the WOOMERA THREAD DIES
			//woomera_set_flag(woomera, WFLAG_HANGUP);
			log_printf(2, woomera->log, "WOOMERA CMD: Hangup Received: [%s]\n", 	woomera->interface);	
			socket_printf(woomera->socket, "200 HANGUP OK%s", WOOMERA_RECORD_SEPERATOR);
			
		} else {
			log_printf(2, woomera->log, "WOOMERA CMD: Hangup NO CHAN SPAN [%s]\n", 	woomera->interface);	
			socket_printf(woomera->socket, "405 NO SUCH CHANNEL%s", WOOMERA_RECORD_SEPERATOR);
		}
    } else if (!strcasecmp(wmsg->command, "call")) {
		char *raw = woomera_message_header(wmsg, "raw-audio");
		call_signal_event_t event;
		char *calling = woomera_message_header(wmsg, "local-number");
		char *called = wmsg->callid;
		char *grp = called;
		char *p;
		int tg = 0;

		if (server.max_calls && server.call_count >= server.max_calls) {
			socket_printf(woomera->socket, "405 Server FULL!%s", WOOMERA_RECORD_SEPERATOR);
			log_printf(2, woomera->log, "SERVER FULL %d/%d\n", server.call_count, server.max_calls);
		} else {
			
			log_printf(2, woomera->log, "New Call %d/%d\n", server.call_count, server.max_calls);

			if ((p = strchr(called, '/'))) {
				*p = '\0';
				called = p+1;
				tg = atoi(grp) - 1;
			}

			if (raw) {
				woomera_set_raw(woomera, raw);
			}
			woomera->index = add_to_holding_tank(woomera);
			socket_printf(woomera->socket, "100 Trying%s", WOOMERA_RECORD_SEPERATOR);

			call_signal_call_init(&event, calling, called, woomera->index);
			event.trunk_group = tg;
			if (call_signal_connection_write(&server.mcon, &event) <= 0) {
				log_printf(0, server.log, "Critical System Error: Failed to tx on ISUP socket [%s]: %s\n", 
						strerror(errno));
			}
		
			log_printf(2, server.log, "Call Called Event ID = %d\n",
					   woomera->index);
		}
    } else if ((media = !strcasecmp(wmsg->command, "media")) || (answer = !strcasecmp(wmsg->command, "answer"))) {
		char *raw = woomera_message_header(wmsg, "raw-audio");
		call_signal_event_t event;
		
		log_printf(2, woomera->log, "WOOMERA CMD: MEDIA [%s]\n", woomera->interface);	
		
		if (woomera_test_flag(woomera, WFLAG_HANGUP) || woomera_test_flag(woomera, WFLAG_MEDIA_END)) {
			struct woomera_event wevent;
			log_printf(0, server.log,"ERROR! call was cancelled MEDIA on HANGUP or MEDIA END!\n");
			socket_printf(woomera->socket, "501 call was cancelled!%s", WOOMERA_RECORD_SEPERATOR);
			new_woomera_event_printf(&wevent, "EVENT HANGUP %s%s",
							 	 wmsg->callid,
								 WOOMERA_RECORD_SEPERATOR
								 );
			enqueue_event(woomera, &wevent);
			woomera->timeout=0;
		} else {
		
			log_printf(0, server.log,"WOOMERA: GOT MEDIA EVENT: [%s]  RAW=%p\n",
						wmsg->callid,raw);
		
			if (raw) {
				
				woomera_set_raw(woomera, raw);
				woomera_set_interface(woomera, wmsg->callid);
					
				if (launch_media_thread_hold_check(woomera)) {
					struct woomera_event wevent;
					log_printf(0, server.log,"ERROR: Failed to Launch Call [%s]\n",
						woomera->interface);
					socket_printf(woomera->socket, "501 call was cancelled!%s", WOOMERA_RECORD_SEPERATOR);
					woomera->timeout=0;

					new_woomera_event_printf(&wevent, "EVENT HANGUP %s%s",
								 	 wmsg->callid,
									 WOOMERA_RECORD_SEPERATOR
									 );
					
					enqueue_event(woomera, &wevent);	
					woomera_set_flag(woomera, WFLAG_HANGUP);
					woomera_set_flag(woomera, WFLAG_MEDIA_END);
					woomera_clear_flag(woomera, WFLAG_RUNNING);
					return;
				}
			}
		
			if (!woomera_test_flag(&server.master_connection, WFLAG_MONITOR_RUNNING)) {
				log_printf(0, server.log,"ERROR! Monitor Thread not running!\n");
			} else {
				if (answer && woomera->ms) {
					pthread_mutex_lock(&woomera->ms_lock);
					time(&woomera->ms->answered);
					call_signal_event_init(&event, SIGBOOST_EVENT_CALL_ANSWERED, woomera->ms->chan, woomera->ms->span);
					if (call_signal_connection_write(&server.mcon, &event) <= 0) {
						log_printf(0, server.log, "Critical System Error: Failed to tx on ISUP socket: %s\n", 
						strerror(errno));
					}

					log_printf(2, server.log,"Sent SIGBOOST_EVENT_CALL_ANSWERED Span=%i Chan=%i [w%dg%d]\n",woomera->ms->span,woomera->ms->chan,
						woomera->ms->span,woomera->ms->chan);
					pthread_mutex_unlock(&woomera->ms_lock);
				}
			}
			socket_printf(woomera->socket, "200 %s %s OK%s", wmsg->callid, answer ? "ANSWER" : "MEDIA", WOOMERA_RECORD_SEPERATOR);
		}
    } else {
		socket_printf(woomera->socket, "501 Command '%s' not implemented%s", wmsg->command, WOOMERA_RECORD_SEPERATOR);
    }
}


/*
  EVENT INCOMING 1
  Remote-Address: 10.3.3.104
  Remote-Number:
  Remote-Name: Anthony Minessale!8668630501
  Protocol: H.323
  User-Agent: Post Increment Woomera		1.0alpha1 (OpenH323 v1.17.2)	9/61
  H323-Call-Id: 887b1ff8-bb1f-da11-85c0-0007e98988c4
  Local-Number: 996
  Start-Time: Fri, 09 Sep 2005 12:25:14 -0400
  Local-Name: root
*/


static void handle_call_answer(call_signal_event_t *event)
{
    struct woomera_interface *woomera = NULL;
	int kill = 0;

    pthread_mutex_lock(&server.process_lock);
    woomera = server.process_table[event->span][event->chan];
    pthread_mutex_unlock(&server.process_lock);	
    
    if (woomera && woomera->raw) {
		char callid[80];
		struct woomera_event wevent;
		
		if (woomera == &woomera_dead_dev || 
		    woomera == &woomera_holding_dev ||
		    woomera == &woomera_dead_timeout_dev) {
			  log_printf(0, server.log, "Critical Error: Woomera Device got special dev!\n"); 
			  return; 
		}

		if (woomera_test_flag(woomera, WFLAG_ANSWER)) {
			log_printf(1, server.log, "Refusing to double-answer a call!\n");
			return;
		}

		woomera_set_flag(woomera, WFLAG_ANSWER);

		if (woomera->span != event->span || woomera->chan != event->chan) {
			log_printf(1, server.log, "Refusing to start media on a different channel from the one we agreed on.!\n");
			kill++;
			return;
		}

		if (woomera_test_flag(woomera, WFLAG_HANGUP)) {
			log_printf(1, server.log, "Refusing to answer a dead call!\n");
			kill++;
		} else {
			int err;
			err=0;
			sprintf(callid, "w%dg%d", event->span + 1, event->chan + 1);
			woomera_set_interface(woomera, callid);
#ifndef WOOMERA_EARLY_MEDIA
			pthread_mutex_lock(&server.process_lock);
			err=launch_media_thread(woomera);
			pthread_mutex_unlock(&server.process_lock);
			if (err) {
				log_printf(0, server.log,"ERROR: Failed to Launch Call [%s]\n",
					woomera->interface);
				socket_printf(woomera->socket, "501 call was cancelled!%s", WOOMERA_RECORD_SEPERATOR);	
				kill++;
			} 
#endif

			if (!kill) {
				new_woomera_event_printf(&wevent, "EVENT CONNECT w%dg%d%s",
									 event->span+1,
									 event->chan+1,
									 WOOMERA_RECORD_SEPERATOR
									 );
				enqueue_event(woomera, &wevent);
			}
		}
        } else {
		log_printf(1, server.log, "Answer requested on non-existant session.\n");
		kill++;
	}

	if (kill) {
		call_signal_event_t oevent;
		
		if (woomera) {
			if (woomera_test_flag(woomera, WFLAG_HANGUP)) {
				/* Do not hangup if this call was already hungup */
				return;
			}
			
			woomera_set_flag(woomera, WFLAG_HANGUP);	
		}

		call_signal_event_init(&oevent, SIGBOOST_EVENT_CALL_STOPPED, event->chan, event->span);
		oevent.release_cause = SIGBOOST_RELEASE_CAUSE_NORMAL;
		
		if (call_signal_connection_write(&server.mcon, &oevent) <= 0) {
			log_printf(0, server.log, "Critical System Error: Failed to tx on ISUP socket: %s\n", 
						strerror(errno));
		}
		log_printf(2, server.log, "Sent Refusal SIGBOOST_EVENT_CALL_STOPPED %d/%d\n", event->span, event->chan);
	}
}

static void handle_call_start_ack(call_signal_event_t *event)
{
    struct woomera_interface *woomera = NULL;
    struct woomera_event wevent;
	int kill = 0;

    if ((woomera = pull_from_holding_tank(event->call_setup_id))) {
		char callid[80];
		
		if (woomera_test_flag(woomera, WFLAG_HANGUP)) {
			log_printf(1, server.log, "Refusing to ack a dead call!\n");
			kill++;
		} else {
			int err;
			sprintf(callid, "w%dg%d", event->span + 1, event->chan + 1);
			woomera_set_interface(woomera, callid);
			
#ifdef WOOMERA_EARLY_MEDIA
			pthread_mutex_lock(&server.process_lock);
			err=launch_media_thread(woomera);
			pthread_mutex_unlock(&server.process_lock);
			if (err) {
				log_printf(0, server.log,"ERROR: Failed to Launch Call [%s]\n",
					woomera->interface);
				socket_printf(woomera->socket, "501 call was cancelled!%s", WOOMERA_RECORD_SEPERATOR);	
				kill++;
			}
#endif
			if (!kill) {
				socket_printf(woomera->socket, "201 Accepted%s", WOOMERA_RECORD_SEPERATOR);
				new_woomera_event_printf(&wevent, "EVENT PROCEED w%dg%d%s",
										event->span+1,
										event->chan+1,
										WOOMERA_RECORD_SEPERATOR
										);
				enqueue_event(woomera, &wevent);
				woomera->span = event->span;
				woomera->chan = event->chan;
	
				log_printf(2, server.log, "Call Answered Event ID = %d  Device = w%dg%d!\n",
						event->call_setup_id,woomera->span+1,woomera->chan+1);
			}
		}
	} else {
		log_printf(1, server.log, "Event referrs to a non-existant session (%d)!\n",
				event->call_setup_id);
		kill++;
	}

	if (kill) {
		call_signal_event_t oevent;

		if (woomera) {
			if (woomera_test_flag(woomera, WFLAG_HANGUP)) {
				/* Do not hangup if this call was already hungup */
				return;
			}
			
			woomera_set_flag(woomera, WFLAG_HANGUP);	
		}

		call_signal_event_init(&oevent, SIGBOOST_EVENT_CALL_STOPPED, event->chan, event->span);
		oevent.release_cause = SIGBOOST_RELEASE_CAUSE_NORMAL;
		
		if (call_signal_connection_write(&server.mcon, &oevent) <= 0) {
			log_printf(0, server.log, "Critical System Error: Failed to tx on ISUP socket: %s\n", 
						strerror(errno));
		}
		log_printf(2, server.log, "Sent Refusal SIGBOOST_EVENT_CALL_STOPPED %d/%d\n", event->span, event->chan);
	}

}

static void handle_call_start_nack(call_signal_event_t *event)
{
    struct woomera_interface *woomera = NULL;

    if ((woomera=pull_from_holding_tank(event->call_setup_id))) {
		struct woomera_event wevent;

		if (woomera_test_flag(woomera, WFLAG_HANGUP)) {
			log_printf(1, server.log, "Refusing to nack a dead call!\n");
		} else {
			socket_printf(woomera->socket, "501 Error!%s", WOOMERA_RECORD_SEPERATOR);
			new_woomera_event_printf(&wevent, "EVENT HANGUP w%dg%d%s",
									 event->span+1,
									 event->chan+1,
									 WOOMERA_RECORD_SEPERATOR
									 );
			enqueue_event(woomera, &wevent);
			woomera_clear_flag(woomera, WFLAG_RUNNING);
			
			/* Do not send CALL STOP after a NACK */
			woomera_set_flag(woomera, WFLAG_HANGUP);
		}
	} else {
		log_printf(1, server.log, "Event referrs to a non-existant session!\n");
	}
}

static void handle_call_start(call_signal_event_t *event)
{
    struct woomera_event wevent;
	
    pthread_mutex_lock(&server.process_lock);
    
    if (server.process_table[event->span][event->chan] == &woomera_dead_timeout_dev) {
    	
	log_printf(0, server.log, "Got Start Call on Timeout Device: Proceeding [w%dg%d]\n", 
		event->span+1, event->chan+1);
    
    } else if (server.process_table[event->span][event->chan]) {
    	call_signal_event_t oevent;
	
	call_signal_event_init(&oevent, SIGBOOST_EVENT_CALL_STOPPED, event->chan, event->span);
	oevent.release_cause = SIGBOOST_RELEASE_CAUSE_BUSY;
	if (call_signal_connection_write(&server.mcon, &oevent) <= 0){
		log_printf(0, server.log, "Critical System Error: Failed to tx on ISUP socket %s\n", 
				strerror(errno));
	}
	
	log_printf(2, server.log, "Sent (From Handle START) Call in Timeout SIGBOOST_EVENT_CALL_STOPPED  [w%dg%d]\n", event->span+1, event->chan+1);
	
	pthread_mutex_unlock(&server.process_lock);
	return;
				
    }	
     
    server.process_table[event->span][event->chan] = &woomera_holding_dev;
    
    pthread_mutex_unlock(&server.process_lock);  
    
    new_woomera_event_printf(&wevent, "EVENT INCOMING w%dg%d%s"
							 "Remote-Number: %s%s"
							 "Remote-Name: %s%s"
							 "Protocol: SS7%s"
							 "User-Agent: sangoma_mgd%s"
							 "Local-Number: %s%s"
							 "Channel-Name: Sangoma-w%dg%d%s"
							 "%s"
							 ,
							 event->span+1,
							 event->chan+1,
							 WOOMERA_LINE_SEPERATOR,
							 event->calling_number_digits,
							 WOOMERA_LINE_SEPERATOR,
							 "",
							 WOOMERA_LINE_SEPERATOR,
							 WOOMERA_LINE_SEPERATOR,
							 WOOMERA_LINE_SEPERATOR,
							 event->called_number_digits,
							 WOOMERA_LINE_SEPERATOR,
							 event->span+1,
							 event->chan+1,
							 WOOMERA_LINE_SEPERATOR,
							 WOOMERA_LINE_SEPERATOR
							 );

    if (enqueue_event_on_listeners(&wevent)) {
		enqueue_event(&server.master_connection, &wevent);
    }
    
       
}

static void handle_call_stop(call_signal_event_t *event)
{
    struct woomera_interface *woomera;
    
    pthread_mutex_lock(&server.process_lock);
    woomera = server.process_table[event->span][event->chan];
    
    if (woomera == &woomera_holding_dev) {
    	log_printf(2, server.log, "Event CALL STOP on HOLDING DEVICE [w%dg%d]!\n",
			event->span+1,event->chan+1);
    	server.process_table[event->span][event->chan]=&woomera_dead_dev;
	pthread_mutex_unlock(&server.process_lock);
	return;
    }
    
    pthread_mutex_unlock(&server.process_lock);
    
        if (woomera) {
	
		if (woomera == &woomera_dead_dev || 
		    woomera == &woomera_holding_dev ||
		    woomera == &woomera_dead_timeout_dev) {
			  log_printf(2, server.log, "Critical Error: Woomera Device got specail dev! (O=%p H=%p D=%p DT=%p)\n",
					 woomera,&woomera_holding_dev,&woomera_dead_dev,&woomera_dead_timeout_dev); 
			  return; 
		}
	
		if (woomera->ms) {
			pthread_mutex_lock(&woomera->ms_lock);
			shutdown(woomera->ms->sangoma_sock, SHUT_RDWR);
			shutdown(woomera->ms->udp_sock, SHUT_RDWR);
			pthread_mutex_unlock(&woomera->ms_lock);
		}
			
		log_printf(2, server.log, "Event CALL STOP on w%dg%d ptr=%p ms=%p\n",
				woomera->span+1,woomera->chan+1,woomera,woomera->ms);
				
		if (woomera->ms) {
			pthread_mutex_lock(&woomera->ms_lock);
			shutdown(woomera->ms->sangoma_sock, SHUT_RDWR);
			shutdown(woomera->ms->udp_sock, SHUT_RDWR);
			pthread_mutex_unlock(&woomera->ms_lock);
		}
		woomera_set_flag(woomera, WFLAG_MEDIA_END);
		/* NENAD: Only hangup after the woomera thread dies */
		/* woomera_set_flag(woomera, WFLAG_HANGUP); */
		
	} else {
		log_printf(2, server.log, "Event referrs to a non-existant session!\n");
	}
}

static void handle_heartbeat(call_signal_event_t *event)
{
	call_signal_event_t oevent;

	if (!woomera_test_flag(&server.master_connection, WFLAG_MONITOR_RUNNING)) {
		log_printf(0, server.log,"ERROR! Monitor Thread not running!\n");
	} else {	
		call_signal_event_init(&oevent, SIGBOOST_EVENT_HEARTBEAT, event->chan, event->span);
		if (call_signal_connection_write(&server.mcon, &oevent) <= 0){
			log_printf(0, server.log, "Critical System Error: Failed to tx on ISUP socket %s\n", 
				strerror(errno));
		}
	}
	return;
}


static void handle_call_stop_ack_timeout(call_signal_event_t *event)
{
    struct woomera_interface *woomera;

    pthread_mutex_lock(&server.process_lock);
    woomera = server.process_table[event->span][event->chan];
    
    if (woomera == &woomera_dead_dev) {
    	log_printf(2, server.log, "Event CALL STOP ACK TIMEOUT on DEAD DEVICE [w%dg%d]!\n",
			event->span+1,event->chan+1);
    	server.process_table[event->span][event->chan]=&woomera_dead_timeout_dev;
	
    } else {
    	log_printf(0, server.log, "Event CALL STOP ACK TIMEOUT on INVALID DEVICE [w%dg%d] (O=%p H=%p D=%p DT=%p)! \n",
			event->span+1,event->chan+1,
			woomera,
			&woomera_holding_dev,
			&woomera_dead_dev,
			&woomera_dead_timeout_dev);
    }
    
    pthread_mutex_unlock(&server.process_lock);

}

static void validate_number(char *s)
{
	char *p;
	for (p = s; *p; p++) {
		if (*p < 48 || *p > 57) {
			log_printf(0, server.log, "Encountered a non-numeric character [%c]!\n", *p);
			*p = '\0';
			break;
		}
	}
}

static int parse_ss7_event(call_signal_event_t *event)
{
    int ret = 0;
	
	validate_number(event->called_number_digits);
	validate_number(event->calling_number_digits);
	
    log_printf(2, server.log,
			   "\nEVENT\n"
			   "===================================\n"
			   "		Type: %s (%0x HEX)\n"
			   "		Span: %d\n"
			   "		Chan: %d\n"
			   "   CalledNum: %s\n"
			   "  CallingNum: %s\n"
			   "  Interface : w%dg%d\n"
			   "   Event ID : %d\n"
			   "===================================\n"
			   "\n",
			   call_signal_event_id_name(event->event_id),
			   event->event_id,
			   event->span+1,
			   event->chan+1,
			   (event->called_number_digits_count ? (char *) event->called_number_digits : "N/A"),
			   (event->calling_number_digits_count ? (char *) event->calling_number_digits : "N/A"),
			   event->span+1,
			   event->chan+1,
			   event->event_id
			   );

	

    switch(event->event_id) {

    case SIGBOOST_EVENT_CALL_START:
		handle_call_start(event);
		break;
    case SIGBOOST_EVENT_CALL_STOPPED:
		handle_call_stop(event);
		break;

    case SIGBOOST_EVENT_CALL_START_ACK:
		handle_call_start_ack(event);
		break;
    case SIGBOOST_EVENT_CALL_START_NACK:
		handle_call_start_nack(event);
		break;
    case SIGBOOST_EVENT_CALL_ANSWERED:
		handle_call_answer(event);
		break;
    case SIGBOOST_EVENT_CALL_STOPPED_ACK_TIMEOUT:
    		handle_call_stop_ack_timeout(event);
		break;
    case SIGBOOST_EVENT_HEARTBEAT:
		handle_heartbeat(event);
		break;
    default:
		log_printf(0, server.log, "Warning no handler implemented for [%s]\n", call_signal_event_id_name(event->event_id));
		break;
    }
	
    return ret;
}


static void *monitor_thread_run(void *obj)
{
    int ss = 0;
    call_signal_event_t event;

    server.thread_count++;

    woomera_set_flag(&server.master_connection, WFLAG_MONITOR_RUNNING);

    if (call_signal_connection_open(&server.mcon, server.boost_local_ip, server.boost_local_port, server.boost_remote_ip, server.boost_remote_port) < 0) {
		log_printf(0, server.log, "FATAL ERROR OPENING UDP SOCKET [%d]\n", server.mcon.socket);
		exit(-1);
    }

    call_signal_event_init(&event, SIGBOOST_EVENT_SYSTEM_RESTART, 0, 0);
    if (call_signal_connection_write(&server.mcon, &event) <= 0){
    	log_printf(0, server.log, "Critical System Error: Failed to tx on ISUP socket: %s\n", 
				strerror(errno));
    }

    log_printf(1, server.log, "Open udp socket [%d]\n", server.mcon.socket);
    log_printf(1, server.log, "Monitor Thread Started\n");
    
    while (woomera_test_flag(&server.master_connection, WFLAG_RUNNING) && woomera_test_flag(&server.master_connection, WFLAG_MONITOR_RUNNING)) {
		ss = waitfor_socket(server.mcon.socket, 1000, POLLERR | POLLIN);
		if (ss > 0) {
			call_signal_event_t *event;

			if ((event = call_signal_connection_read(&server.mcon))) {
				log_printf(1, server.log, "Socket Event [%s]!\n", call_signal_event_id_name(event->event_id));
				parse_ss7_event(event);
			}
		}else if (ss < 0){
			log_printf(0, server.log, "Thread Run: Select Socket Error!\n");
			break;
		}
    }

    log_printf(1, server.log, "Close udp socket [%d]\n", server.mcon.socket);
    call_signal_connection_close(&server.mcon);
    server.thread_count--;
    woomera_clear_flag(&server.master_connection, WFLAG_MONITOR_RUNNING);
    log_printf(1, server.log, "Monitor Thread Ended\n");
    return NULL;
}

static void *woomera_thread_run(void *obj)
{
    struct woomera_interface *woomera = obj;
    struct woomera_message wmsg;
    struct woomera_event wevent;
    char *event_string;
    int mwi;

    log_printf(1, server.log, "WOOMERA session for started (ptr=%p)\n", 
    			woomera);
    
    server.thread_count++;
    woomera_set_flag(woomera, WFLAG_RUNNING);
    pthread_mutex_init(&woomera->queue_lock, NULL);	
    pthread_mutex_init(&woomera->ms_lock, NULL);	

    socket_printf(woomera->socket,
				  "EVENT HELLO 0 1.0 Anthony Minessale%s"
				  "Supported-Protocols: TDM%s"
				  "Version: 0.1%s"
				  "Remote-Address: %s%s"
				  "Remote-Port: %d%s",
				  WOOMERA_LINE_SEPERATOR,
				  WOOMERA_LINE_SEPERATOR,
				  WOOMERA_LINE_SEPERATOR,
				  inet_ntoa(woomera->addr.sin_addr), WOOMERA_LINE_SEPERATOR,
				  ntohs(woomera->addr.sin_port), WOOMERA_RECORD_SEPERATOR
				  );


    woomera_set_interface(woomera, inet_ntoa(woomera->addr.sin_addr));
	
    while (woomera_test_flag(&server.master_connection, WFLAG_RUNNING) &&
	   woomera_test_flag(woomera, WFLAG_RUNNING) &&
	   !woomera_test_flag(woomera, WFLAG_HANGUP) ) {
	   
	   mwi = woomera_message_parse(woomera, &wmsg, WOOMERA_HARD_TIMEOUT);
	   if (mwi >= 0) {

		if (mwi) {
			interpret_command(woomera, &wmsg);			
		} else if (woomera_test_flag(woomera, WFLAG_EVENT)){
			while ((event_string = dequeue_event(woomera))) {
				socket_printf(woomera->socket, "%s", event_string);
				free(event_string);
			}
			woomera_clear_flag(woomera, WFLAG_EVENT);
		}

		if (woomera->timeout > 0 && time(NULL) >= woomera->timeout) {
		
			new_woomera_event_printf(&wevent, "EVENT HANGUP %s%sTimeout: %ld%s",
									 woomera->interface,
									 WOOMERA_LINE_SEPERATOR,
									 woomera->timeout,
									 WOOMERA_RECORD_SEPERATOR
									 );
			enqueue_event(woomera, &wevent);

			log_printf(2, server.log, "WOOMERA session (ptr=%p) Call Timedout ! [%s]\n",
					 woomera,woomera->interface);
		
			/* NENAD Let the Index check run a nak */
			woomera_set_flag(woomera, WFLAG_HANGUP);
			break;
		
		}
	  } else  {
	  	 log_printf(1, server.log, "WOOMERA session (ptr=%p) [%s] READ MSG Error %i \n", 
    			woomera,woomera->interface,mwi);
		 break;
	  }
		
    }
    
    log_printf(1, server.log, "WOOMERA session (ptr=%p) is dying [%s]:  SR=%d WR=%d WF=0x%04X\n", 
    			woomera,woomera->interface,
			woomera_test_flag(&server.master_connection, WFLAG_RUNNING),
			woomera_test_flag(woomera, WFLAG_RUNNING),
			woomera->flags);

    if (woomera_test_flag(woomera, WFLAG_MEDIA_RUNNING) && woomera->ms) {
		woomera_set_flag(woomera, WFLAG_MEDIA_END);
		while(woomera_test_flag(woomera, WFLAG_MEDIA_RUNNING)) {
			usleep(1000);
			sched_yield();
		}
    }

	if (!woomera_test_flag(woomera, WFLAG_HANGUP)) {
		call_signal_event_t event;
		woomera_set_flag(woomera, WFLAG_HANGUP);
		int span = -1, chan = -1;
		
		chan = woomera->chan;
		span = woomera->span;
		
		if (chan < 0 || span < 0) {
			sscanf(woomera->interface, "w%dg%d",  &span, &chan);
			span--;
			chan--;
		}

		if (span > -1 && chan > -1) {
			
			pthread_mutex_lock(&server.process_lock);
			/* This is possible in case media thread dies on startup */
    			if (server.process_table[span][chan] == woomera){
				server.process_table[span][chan] = NULL;
			}
    			pthread_mutex_unlock(&server.process_lock);	
			
			call_signal_event_init(&event, SIGBOOST_EVENT_CALL_STOPPED, chan, span);
			event.release_cause = string_to_release(woomera->cause);

			if (call_signal_connection_write(&server.mcon, &event) <= 0) {
				log_printf(0, server.log, "Critical System Error: Failed to tx on ISUP socket: %s\n", 
						strerror(errno));
			}
			
			log_printf(2, woomera->log, "Sent (Ack) SIGBOOST_EVENT_CALL_STOPPED %d/%d [%s] ptr=%p\n", 
						span, chan,woomera->interface,woomera);
		
		}else{
			log_printf(2, woomera->log, "FAILED: Sent (ERROR) SIGBOOST_EVENT_CALL_STOPPED %d/%d [%s] Index=%d ptr=%p\n", span, chan, woomera->interface, woomera->index, woomera);
		}
	}
	
	if (woomera->index) { 
		call_signal_event_t event;
		int index = woomera->index;
		pull_from_holding_tank(index);

		call_signal_event_init(&event, SIGBOOST_EVENT_CALL_START_NACK, 0, 0);
		event.call_setup_id = index;
		if (call_signal_connection_write(&server.mcon, &event) <=0) {
			log_printf(0, server.log, "Critical System Error: Failed to tx on ISUP socket: %s\n", 
						strerror(errno));
		}
		log_printf(2, woomera->log, "Sent SIGBOOST_EVENT_CALL_START_NACK %d\n", index);
	}

    if (woomera_test_flag(woomera, WFLAG_EVENT)){
		while ((event_string = dequeue_event(woomera))) {
			socket_printf(woomera->socket, "%s", event_string);
			free(event_string);
		}
		woomera_clear_flag(woomera, WFLAG_EVENT);
    }

    usleep(3000000);
    log_printf(1, woomera->log, "Thread Finished %u\n", (unsigned long) woomera->thread);
    close_socket(&woomera->socket);
    woomera->socket=-1;

    if (woomera_test_flag(woomera, WFLAG_LISTENING)) {
		del_listener(woomera);
    }

    /* delete queue */
    while ((event_string = dequeue_event(woomera))) {
		free(event_string);
    }

    log_printf(1, server.log, "WOOMERA session for [%s] stopped (ptr=%p)\n", 
    			woomera->interface,woomera);
    
    pthread_mutex_destroy(&woomera->queue_lock);
    pthread_mutex_destroy(&woomera->ms_lock);
    woomera_set_raw(woomera, NULL);
    woomera_set_interface(woomera, NULL);
    woomera_set_cause(woomera, NULL);
    server.call_count--;


    free(woomera);
    server.thread_count--;
    return NULL;
}


static void launch_woomera_thread(struct woomera_interface *woomera) 
{
    pthread_attr_t attr;
    int result = 0;

    result = pthread_attr_init(&attr);
    pthread_attr_setschedpolicy(&attr, SCHED_RR);
    pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_DETACHED);
	pthread_attr_setstacksize(&attr, MGD_STACK_SIZE);
    result = pthread_create(&woomera->thread, &attr, woomera_thread_run, woomera);
    result = pthread_attr_destroy(&attr);
}


static void launch_monitor_thread(void) 
{
    pthread_attr_t attr;
    int result = 0;

    result = pthread_attr_init(&attr);
    pthread_attr_setschedpolicy(&attr, SCHED_RR);
    pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_DETACHED);
	pthread_attr_setstacksize(&attr, MGD_STACK_SIZE);
    result = pthread_create(&server.monitor_thread, &attr, monitor_thread_run, NULL);
    result = pthread_attr_destroy(&attr);
}

static int configure_server(void)
{
    struct woomera_config cfg;
    char *var, *val;
    int cnt = 0;

    if (!woomera_open_file(&cfg, server.config_file)) {
		log_printf(0, server.log, "open of %s failed\n", server.config_file);
		return 0;
    }
	
    while (woomera_next_pair(&cfg, &var, &val)) {
		if (!strcasecmp(var, "boost_local_ip")) {
			strncpy(server.boost_local_ip, val, sizeof(server.boost_local_ip) -1);
			cnt++;
		} else if (!strcasecmp(var, "boost_local_port")) {
			server.boost_local_port = atoi(val);
			cnt++;
		} else if (!strcasecmp(var, "boost_remote_ip")) {
			strncpy(server.boost_remote_ip, val, sizeof(server.boost_remote_ip) -1);
			cnt++;
		} else if (!strcasecmp(var, "boost_remote_port")) {
			server.boost_remote_port = atoi(val);
			cnt++;
		} else if (!strcasecmp(var, "logfile_path")) {
			if (!server.logfile_path) {
				server.logfile_path = strdup(val);
			}
		} else if (!strcasecmp(var, "woomera_port")) {
			server.port = atoi(val);
		} else if (!strcasecmp(var, "debug_level")) {
			server.debug = atoi(val);
		} else if (!strcasecmp(var, "max_calls")) {
			int max = atoi(val);
			if (max > 0) {
				server.max_calls = max;
			}
		} else if (!strcasecmp(var, "media_ip")) {
			strncpy(server.media_ip, val, sizeof(server.media_ip) -1);
		} else {
			log_printf(0, server.log, "Invalid Option %s at line %d!\n", var, cfg.lineno);
		}
    }

    woomera_close_file(&cfg);
    return cnt == 4 ? 1 : 0;
}


static int main_thread(void)
{
    struct sockaddr_in sock_addr, client_addr;
    struct woomera_interface *new_woomera;
    int client_sock = -1, pid = 0;
    int len = 0;
    FILE *tmp;

    if ((server.master_connection.socket = socket(PF_INET, SOCK_STREAM, IPPROTO_TCP)) < 0) {
		log_printf(0, server.log, "socket() failed\n");
		return 1;
    }
	
    memset(&sock_addr, 0, sizeof(sock_addr));	/* Zero out structure */
    sock_addr.sin_family = AF_INET;				   /* Internet address family */
    sock_addr.sin_addr.s_addr = htonl(INADDR_ANY); /* Any incoming interface */
    sock_addr.sin_port = htons(server.port);	  /* Local port */

    /* Bind to the local address */
    if (bind(server.master_connection.socket, (struct sockaddr *) &sock_addr, sizeof(sock_addr)) < 0) {
		log_printf(0, server.log, "bind(%d) failed\n", server.port);
		return 1;
    }
	 
    /* Mark the socket so it will listen for incoming connections */
    if (listen(server.master_connection.socket, MAXPENDING) < 0) {
		log_printf(0, server.log, "listen() failed\n");
		return 1;
    }

    if ((pid = get_pid_from_file(PIDFILE))) {
		log_printf(0, stderr, "pid %d already exists.\n", pid);
		exit(0);
    }

    if (!(tmp = safe_fopen(PIDFILE, "w"))) {
		log_printf(0, stderr, "Error creating pidfile %s\n", PIDFILE);
		return 1;
    } else {
		fprintf(tmp, "%d", getpid());
		fclose(tmp);
		tmp = NULL;
    }

    no_nagle(server.master_connection.socket);

    log_printf(1, server.log, "Woomera Ready port: %d\n", server.port);
    while (woomera_test_flag(&server.master_connection, WFLAG_RUNNING)) {
		/* Set the size of the in-out parameter */
		len = sizeof(client_addr);

		/* Wait for a client to connect */
		if ((client_sock = accept(server.master_connection.socket, (struct sockaddr *) &client_addr, &len)) < 0) {
			log_printf(0, server.log, "accpet() failed\n");
			return 1;
		}

		if ((new_woomera = new_woomera_interface(client_sock, &client_addr, len))) {
			log_printf(0, server.log, "Starting Thread for New Connection %s:%d Sock=%d\n",
					   inet_ntoa(new_woomera->addr.sin_addr),
					   ntohs(new_woomera->addr.sin_port),
					   client_sock);
			launch_woomera_thread(new_woomera);
		} else {
			log_printf(0, server.log, "Critical ERROR: memory/socket error\n");
		}
    }

    return 0;
}

static int do_ignore(int sig)
{
    return 0;
}

static int do_shut(int sig)
{
    woomera_clear_flag(&server.master_connection, WFLAG_RUNNING);
    close_socket(&server.master_connection.socket);
    log_printf(1, server.log, "Caught SIG %d, Closing Master Socket!\n", sig);
    return 0;
}

static int woomera_startup(int argc, char **argv)
{
    int x = 0, pid = 0, bg = 0;
    char *cfg=NULL, *debug=NULL, *arg=NULL;

    while((arg = argv[x++])) {

		if (!strcasecmp(arg, "-hup")) {
			if (! (pid = get_pid_from_file(PIDFILE))) {
				log_printf(0, stderr, "Error reading pidfile %s\n", PIDFILE);
				exit(1);
			} else {
				log_printf(0, stderr, "Killing PID %d\n", pid);
				kill(pid, SIGHUP);
				sleep(1);
				exit(0);
			}
		
		 } else if (!strcasecmp(arg, "-term")) {
                        if (! (pid = get_pid_from_file(PIDFILE))) {
                                log_printf(0, stderr, "Error reading pidfile %s\n", PIDFILE);
                                exit(1);
                        } else {
                                log_printf(0, stderr, "Killing PID %d\n", pid);
                                kill(pid, SIGTERM);
                                unlink(PIDFILE);
				sleep(1);
                                exit(0);
                        }	
			
		} else if (!strcasecmp(arg, "-help")) {
			fprintf(stdout, "%s\n%s [-help] | [-hup] | [-wipe] | [[-bg] | [-debug <level>] | [-cfg <path>] | [-log <path>]]\n\n", WELCOME_TEXT, argv[0]);
			exit(0);
		} else if (!strcasecmp(arg, "-wipe")) {
			unlink(PIDFILE);
		} else if (!strcasecmp(arg, "-bg")) {
			bg = 1;
		} else if (!strcasecmp(arg, "-debug")) {
			if (argv[x] && *(argv[x]) != '-') {
				debug = argv[x++];
			}
		} else if (!strcasecmp(arg, "-cfg")) {
			if (argv[x] && *(argv[x]) != '-') {
				cfg = argv[x++];
			}
		} else if (!strcasecmp(arg, "-log")) {
			if (argv[x] && *(argv[x]) != '-') {
				server.logfile_path = strdup(argv[x++]);
			}
		} else if (*arg == '-') {
			log_printf(0, stderr, "Unknown Option %s\n", arg);
			fprintf(stdout, "%s\n%s [-help] | [-hup] | [-wipe] | [[-bg] | [-debug <level>] | [-cfg <path>] | [-log <path>]]\n\n", WELCOME_TEXT, argv[0]);
			exit(1);
		}
    }

    if (bg && (pid = fork())) {
		log_printf(0, stderr, "Backgrounding!\n");
		return 0;
    }

    server.port = 42420;
    server.debug = 0;
    strcpy(server.media_ip, "127.0.0.1");
    server.next_media_port = WOOMERA_MIN_MEDIA_PORT;
    server.log = stdout;
    server.master_connection.socket = -1;
    server.master_connection.event_queue = NULL;
    server.config_file = cfg ? cfg : "/etc/sangoma_mgd.conf";
    pthread_mutex_init(&server.listen_lock, NULL);	
    pthread_mutex_init(&server.ht_lock, NULL);	
    pthread_mutex_init(&server.process_lock, NULL);
    pthread_mutex_init(&server.media_udp_port_lock, NULL);	
    pthread_mutex_init(&server.master_connection.queue_lock, NULL);	
	server.master_connection.chan = -1;
	server.master_connection.span = -1;

    if (!configure_server()) {
		log_printf(0, server.log, "configuration failed!\n");
		return 0;
    }

    if (server.logfile_path) {
		if (!(server.log = safe_fopen(server.logfile_path, "a"))) {
			log_printf(0, stderr, "Error setting logfile %s!\n", server.logfile_path);
			server.log = stderr;
			return 0;
		}
    }
	
	
    if (debug) {
		server.debug = atoi(debug);
    }


    (void) signal(SIGINT,(void *) do_shut);
    (void) signal(SIGPIPE,(void *) do_ignore);
    (void) signal(SIGHUP,(void *) do_shut);

    woomera_set_flag(&server.master_connection, WFLAG_RUNNING);
    launch_monitor_thread();
    fprintf(stderr, "%s", WELCOME_TEXT);
    log_printf(0, stderr, "Woomera STARTUP Complete.\n");



    return 1;
}

static int woomera_shutdown(void) 
{
    char *event_string;
    int told = 0, loops = 0;

    close_socket(&server.master_connection.socket);
    pthread_mutex_destroy(&server.listen_lock);
    pthread_mutex_destroy(&server.ht_lock);	
    pthread_mutex_destroy(&server.process_lock);
    pthread_mutex_destroy(&server.media_udp_port_lock);
    pthread_mutex_destroy(&server.master_connection.queue_lock);
    woomera_clear_flag(&server.master_connection, WFLAG_RUNNING);


    if (server.logfile_path) {
		free(server.logfile_path);
		server.logfile_path = NULL;
    }

    /* delete queue */
    while ((event_string = dequeue_event(&server.master_connection))) {
		free(event_string);
    }

    while(server.thread_count > 0) {
		loops++;

		if (loops % 1000 == 0) {
			told = 0;
		}

		if (loops > 10000) {
			log_printf(0, server.log, "Red Alert! threads did not stop\n");
			assert(server.thread_count == 0);
		}

		if (told != server.thread_count) {
			log_printf(1, server.log, "Waiting For %d thread%s.\n", server.thread_count, server.thread_count == 1 ? "" : "s");
			told = server.thread_count;
		}
		ysleep(10000);
    }
    unlink(PIDFILE);
    log_printf(0, stderr, "Woomera SHUTDOWN Complete.\n");
    return 0;
}

int main(int argc, char *argv[]) 
{
    int ret = 0;
    
    /* Initialize a holding device */
    memset(&woomera_holding_dev, 0, sizeof(woomera_holding_dev));
    memset(&woomera_dead_dev, 0, sizeof(woomera_dead_dev));
    memset(&woomera_dead_timeout_dev, 0, sizeof(woomera_dead_timeout_dev));
    
    if (! (ret = woomera_startup(argc, argv))) {
		exit(0);
    }
    ret = main_thread();
    woomera_shutdown();
    return ret;
}

/** EMACS **
 * Local variables:
 * mode: C
 * c-basic-offset: 4
 * End:
 */

